#include <ros/ros.h>
#include <ros/callback_queue.h>
#include <cv_bridge/CvBridge.h>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/image_encodings.h>
#include <sensor_msgs/CameraInfo.h>
#include <image_geometry/pinhole_camera_model.h>
#include <message_filters/subscriber.h>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <pcl_ros/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl/PointIndices.h>
#include <Eigen/Core>
#include <ctime>
#include <cstdlib>

using namespace cv;
using namespace std;
namespace enc = sensor_msgs::image_encodings;

typedef pcl::PointCloud<pcl::PointXYZRGB> PointCloud;
typedef union{
	struct
	{
		unsigned char Blue;
		unsigned char Green;
		unsigned char Red;
		unsigned char Alpha;
	};
	float float_value;
	long long_value;
} RGBValue;

IplImage* ir_image = NULL;
IplImage* rgb_image = NULL;
IplImage* depth_image = NULL;
IplImage* reg_depth_image = NULL;
IplImage* ir_img_filtered = NULL;
IplImage* rgb_img_filtered = NULL;

Mat map11, map12, map21, map22;
CvMat *M1, *M2, *D1, *D2, *P1, *P2, *R, *T, *Q;

int img_index = 1; 

class cross_modal_kinect
{
  public:       
        ros::NodeHandle nh;

	ros::Publisher pub_point_cloud_refined_, pub_point_cloud_origin_;
	ros::Publisher pub_ir_, pub_rgb_, pub_ir_filtered_, pub_rgb_filtered_, pub_depth_, pub_depth_refined_, pub_disp_;
 	

        message_filters::Subscriber<sensor_msgs::Image> sub_ir, sub_rgb, sub_depth;
        message_filters::Subscriber<sensor_msgs::CameraInfo> sub_ir_info, sub_rgb_info, sub_depth_info;
        sensor_msgs::CvBridge bridge_;
	sensor_msgs::CameraInfo _camInfoMsg;
	image_geometry::PinholeCameraModel depth_model_, rgb_model_, ir_model_;
	ros::CallbackQueue cq_ir, cq_rgb, cq_depth, cq_ir_info, cq_rgb_info, cq_depth_info;


        cross_modal_kinect(ros::NodeHandle& _nh): nh(_nh)
        {
		pub_point_cloud_refined_ = nh.advertise<PointCloud>("points_refined", 1);
		//pub_point_cloud_origin_ = nh.advertise<PointCloud>("points_origin", 1);
		pub_ir_ = nh.advertise<sensor_msgs::Image>("ir", 1);
		pub_rgb_ = nh.advertise<sensor_msgs::Image>("rgb", 1);
		pub_ir_filtered_ = nh.advertise<sensor_msgs::Image>("ir_filtered", 1);
		pub_rgb_filtered_ = nh.advertise<sensor_msgs::Image>("rgb_filtered", 1);
		pub_depth_ = nh.advertise<sensor_msgs::Image>("depth", 1);
		pub_depth_refined_ = nh.advertise<sensor_msgs::Image>("depth_refined", 1);
		pub_disp_ = nh.advertise<sensor_msgs::Image>("disp", 1);

		sub_ir.subscribe(nh, "ir/image_raw" , 1, ros::TransportHints(), &cq_ir);
                sub_ir.registerCallback(&cross_modal_kinect::sub_ir_callback, this);
		sub_ir.unsubscribe();

                sub_rgb.subscribe(nh, "rgb/image_color" , 1, ros::TransportHints(), &cq_rgb);
                sub_rgb.registerCallback(&cross_modal_kinect::sub_rgb_callback, this);
		sub_rgb.unsubscribe();

		sub_depth.subscribe(nh, "depth/image_raw" , 1, ros::TransportHints(), &cq_depth);
                sub_depth.registerCallback(&cross_modal_kinect::sub_depth_callback, this);
		sub_depth.unsubscribe();
		
		sub_ir_info.subscribe(nh, "ir/camera_info", 1, ros::TransportHints(), &cq_ir_info);
		sub_ir_info.registerCallback(boost::bind(&cross_modal_kinect::caminfo_cb, this, _1, "ir"));
		sub_ir_info.unsubscribe();

		sub_rgb_info.subscribe(nh, "rgb/camera_info", 1, ros::TransportHints(), &cq_rgb_info);
		sub_rgb_info.registerCallback(boost::bind(&cross_modal_kinect::caminfo_cb, this, _1, "rgb"));
		sub_rgb_info.unsubscribe();

		sub_depth_info.subscribe(nh, "depth/camera_info", 1, ros::TransportHints(), &cq_depth_info);
		sub_depth_info.registerCallback(boost::bind(&cross_modal_kinect::caminfo_cb, this, _1, "depth"));
        }

        void sub_ir_callback(const sensor_msgs::Image::ConstPtr& msg)
        {
		IplImage *cv_image = NULL; 
		printf("sub_ir_callback -> ");
    		try
    		{
      			cv_image = bridge_.imgMsgToCv(msg, "mono16");
			ir_image = cvCloneImage(cv_image);
    		}
  		catch (sensor_msgs::CvBridgeException& ex) {
         		ROS_ERROR("[draw_frames] Failed to convert image");
         		return;
    		}
    		IplImage* ir_img_tmp = cvCreateImage( cvSize(ir_image->width, ir_image->height), IPL_DEPTH_8U , 1 );	
    		cvConvertScale(ir_image,ir_img_tmp);
		char filename[255];
		sprintf(filename,"ir-%d.png",img_index-1);
		cv::imwrite(filename,ir_img_tmp);	
    		cvReleaseImage(&ir_img_tmp);
		char raw_filename[255];
		sprintf(raw_filename,"ir-%d.raw",img_index-1);
		FILE *pFile=fopen(raw_filename,"w");
        	fwrite (ir_image->imageData , sizeof(unsigned short) , 640*480 , pFile );
        	fclose (pFile);
		cout << endl << "IR \t\t\t" << msg->header.stamp <<endl;
		sub_rgb.subscribe();
		sub_rgb_info.subscribe();
        }
        void sub_rgb_callback(const sensor_msgs::Image::ConstPtr& msg)
        {
		printf("sub_rgb_callback -> ");
		try{
			rgb_image = bridge_.imgMsgToCv(msg, "bgr8");
    		}
		catch (sensor_msgs::CvBridgeException& ex) {
         		ROS_ERROR("[draw_frames] Failed to convert image");
         		return;
    		}
		char filename[255];
		sprintf(filename,"rgb-%d.png",img_index);
		cv::imwrite(filename,rgb_image);	
                cout << endl << "RGB \t\t\t" << msg->header.stamp <<endl;
		sub_ir.subscribe();
		sub_ir_info.subscribe();
        }
        void sub_depth_callback(const sensor_msgs::Image::ConstPtr& msg)
   	{
		IplImage *cv_image = NULL;
		printf("sub_depth_callback -> ");
    		try
    		{
      			cv_image = bridge_.imgMsgToCv(msg, "mono16");
			depth_image = cvCloneImage(cv_image);
    		}
  		catch (sensor_msgs::CvBridgeException& ex) {
         		ROS_ERROR("[draw_frames] Failed to convert image");
         		return;
    		}
		char raw_filename[255];
		sprintf(raw_filename,"depth-%d.raw",img_index);
		FILE *pFile=fopen(raw_filename,"w");
        	fwrite (depth_image->imageData , sizeof(unsigned short) , 640*480 , pFile );
        	fclose (pFile);
		cout << endl << "DEPTH \t\t\t" << msg->header.stamp <<endl;
	}

	void caminfo_cb(const sensor_msgs::CameraInfoConstPtr &msg, const char* topic)
	{
		_camInfoMsg = *msg;
		/*printf("%s-camera information:\n",topic);		
		printf("width:%d height:%d\n",_camInfoMsg.width, _camInfoMsg.height);
		printf("distortion model:%s\n",_camInfoMsg.distortion_model.c_str());
		printf("distortion paras: [%f %f %f %f %f]\n",_camInfoMsg.D[0], _camInfoMsg.D[1], _camInfoMsg.D[2], _camInfoMsg.D[3], _camInfoMsg.D[4]);
		printf("camera matrix: [%f %f %f]\n",_camInfoMsg.K[0], _camInfoMsg.K[1], _camInfoMsg.K[2]);
		printf("camera matrix: [%f %f %f]\n",_camInfoMsg.K[3], _camInfoMsg.K[4], _camInfoMsg.K[5]);
		printf("camera matrix: [%f %f %f]\n",_camInfoMsg.K[6], _camInfoMsg.K[7], _camInfoMsg.K[8]);
		printf("projection matrix: [%f %f %f %f]\n",_camInfoMsg.P[0], _camInfoMsg.P[1], _camInfoMsg.P[2], _camInfoMsg.P[3]);
		printf("projection matrix: [%f %f %f %f]\n",_camInfoMsg.P[4], _camInfoMsg.P[5], _camInfoMsg.P[6], _camInfoMsg.P[7]);
		printf("projection matrix: [%f %f %f %f]\n",_camInfoMsg.P[8], _camInfoMsg.P[9], _camInfoMsg.P[10], _camInfoMsg.P[11]);x
		if(strcmp(topic,"ir")==0){
			ir_model_.fromCameraInfo(msg);
			printf("ir_mode specified\n");
		}else if(strcmp(topic,"rgb")==0){
			rgb_model_.fromCameraInfo(msg);
			printf("rgb_mode specified\n");
		}else if(strcmp(topic,"depth")==0){
			depth_model_.fromCameraInfo(msg);
			printf("depth_mode specified\n");
		}*/
	}

        ~cross_modal_kinect()
        {
                cout << endl;
        }
};

void calculate_transform(){
	XmlRpc::XmlRpcValue camera_matrix_ir, camera_matrix_rgb, distort_ir, distort_rgb, matrix_R, matrix_T;
	int M_cols, M_rows, D_cols, D_rows, R_cols, R_rows, T_cols, T_rows;
	ros::param::get("/calibration_ir/camera_matrix/data", camera_matrix_ir);
	ros::param::get("/calibration_rgb/camera_matrix/data", camera_matrix_rgb);
	ros::param::get("/calibration_ir/distortion_coefficients/data", distort_ir);
	ros::param::get("/calibration_rgb/distortion_coefficients/data", distort_rgb);
	ros::param::get("/calibration_rgb/camera_matrix/cols", M_cols);
	ros::param::get("/calibration_rgb/camera_matrix/rows", M_rows);
	ros::param::get("/calibration_rgb/distortion_coefficients/cols", D_cols);
	ros::param::get("/calibration_rgb/distortion_coefficients/rows", D_rows);
	ros::param::get("/calibration_ex/R/data", matrix_R);
	ros::param::get("/calibration_ex/R/cols", R_cols);
	ros::param::get("/calibration_ex/R/rows", R_rows);
	ros::param::get("/calibration_ex/T/data", matrix_T);
	ros::param::get("/calibration_ex/T/cols", T_cols);
	ros::param::get("/calibration_ex/T/rows", T_rows);

	M1 = cvCreateMat(M_rows, M_cols, CV_64FC1);
	M2 = cvCreateMat(M_rows, M_cols, CV_64FC1);
	D1 = cvCreateMat(D_rows, D_cols, CV_64FC1);
	D2 = cvCreateMat(D_rows, D_cols, CV_64FC1);
	R = cvCreateMat(R_rows, R_cols, CV_64FC1);
	T = cvCreateMat(T_rows, T_cols, CV_64FC1);
	for (int i = 0; i < camera_matrix_ir.size(); ++i){
		cvSet2D(M1,i/M_cols,i%M_cols,cvRealScalar(static_cast<double>(camera_matrix_ir[i])));
		cvSet2D(M2,i/M_cols,i%M_cols,cvRealScalar(static_cast<double>(camera_matrix_rgb[i])));
	}
	for (int i = 0; i < distort_ir.size(); ++i){
		cvSet1D(D1,i,cvRealScalar(static_cast<double>(distort_ir[i])));
		cvSet1D(D2,i,cvRealScalar(static_cast<double>(distort_rgb[i])));
	}
	for (int i = 0; i < matrix_R.size(); ++i)
		cvSet2D(R,i/R_cols,i%R_cols,cvRealScalar(static_cast<double>(matrix_R[i])));
	for (int i = 0; i < matrix_T.size(); ++i)
		cvSet1D(T,i,cvRealScalar(static_cast<double>(matrix_T[i])));

	cv::Mat M1_tmp(M1,0); cv::Mat M2_tmp(M2,0);
	cv::Mat D1_tmp(D1,0); cv::Mat D2_tmp(D2,0);
	cv::Mat R_tmp(R,0); cv::Mat T_tmp(T,0);
	cv::Mat R1, P1_tmp, R2, P2_tmp, Q_tmp;
	Rect roi1, roi2;
	stereoRectify( M1_tmp, D1_tmp, M2_tmp, D2_tmp, cvSize(640,480), R_tmp, T_tmp, R1, R2, P1_tmp, P2_tmp, Q_tmp, -1, cvSize(640,480), &roi1, &roi2 );
	initUndistortRectifyMap(M1, D1, R1, P1_tmp, cvSize(640,480), CV_16SC2, map11, map12);
	initUndistortRectifyMap(M2, D2, R2, P2_tmp, cvSize(640,480), CV_16SC2, map21, map22);
	P1 = cvCreateMat(3, 4, CV_64FC1);
	P2 = cvCreateMat(3, 4, CV_64FC1);
	for(int i = 0;i < P1->height; i++){
		for(int j = 0;j < P1->width; j++){
			double pp = P1_tmp.at<double>(i,j); cvSet2D(P1,i,j,cvRealScalar(pp));
			double qq = P2_tmp.at<double>(i,j); cvSet2D(P2,i,j,cvRealScalar(qq));
		}
	}
	Q = cvCreateMat(4, 4, CV_64FC1);
	for(int i = 0;i < Q->height; i++){
		for(int j = 0;j < Q->width; j++){
			double pp = Q_tmp.at<double>(i,j);
			cvSet2D(Q,i,j,cvRealScalar(pp));//printf("%f ",pp);
		}
	}
}

//cv::Mat calculate_disparity(IplImage* ir_img_filtered, IplImage* rgb_img_filtered)
IplImage* calculate_disparity(IplImage* ir_img_filtered, IplImage* rgb_img_filtered)
{
	cv::Mat img1(ir_img_filtered,0);
	cv::Mat img2(rgb_img_filtered,0);
	cv::Mat img1r, img2r;
	remap(img1, img1r, map11, map12, INTER_LINEAR);
	remap(img2, img2r, map21, map22, INTER_LINEAR);

	
	StereoSGBM sgbm;
	sgbm.preFilterCap = 63;
	sgbm.SADWindowSize = 5; //blocksize=5
	sgbm.P1 = 8*img1.channels()*sgbm.SADWindowSize*sgbm.SADWindowSize;
	sgbm.P2 = 32*img1.channels()*sgbm.SADWindowSize*sgbm.SADWindowSize;
	sgbm.minDisparity = 0;
	sgbm.numberOfDisparities = 16; //max-disparity=16
	sgbm.uniquenessRatio = 40;
	sgbm.speckleWindowSize = 100;
	sgbm.speckleRange = 32;
	sgbm.disp12MaxDiff = 1;
	sgbm.fullDP = 0;
	Mat disp, disp8;
	sgbm(img1r, img2r, disp);
	disp.convertTo(disp8, CV_8U, 255/(sgbm.numberOfDisparities*16.));

	IplImage* disp8_unrect = cvCreateImage(cvSize(map11.cols,map11.rows), IPL_DEPTH_8U, 1);
	for(int i = 0;i<map11.rows;i++){
		for(int j = 0;j<map11.cols;j++){
			Vec2s point = map11.at<Vec2s>(i, j);
			if(point[0]>0 && point[1]>0 && point[1]<map11.rows && point[0]<map11.cols){
				unsigned short disp_point = disp8.at<uchar>(i, j);
				cvSet2D(disp8_unrect,int(point[1]),int(point[0]),cvScalar(disp_point));
			}
		}
    	}
	char raw_filename[255];
	sprintf(raw_filename,"disp-%d.raw",img_index-1);
	FILE *pFile=fopen(raw_filename,"w");
        fwrite (disp8_unrect->imageData , sizeof(unsigned short) , 640*480 , pFile );
        fclose (pFile);
	char filename[255];
	sprintf(filename,"disp-%d.png",img_index-1);
	cv::imwrite(filename,disp8_unrect);
	imshow("disp", disp);
	imshow("disp8", disp8);
	imshow("rgb_reg", img2r);
 	imshow("ir_reg", img1r);
	cvNamedWindow("DISP UNRECT",1); 
	cvShowImage("DISP UNRECT",disp8_unrect); 
	//waitKey();

	/*cv::Mat depth_image_src(depth_image,0);
	cv::Mat depth_image_dst;
	remap(depth_image_src, depth_image_dst, map11, map12, INTER_LINEAR);
	cvNamedWindow("RGB FILTERED",1); 
	cvShowImage("RGB FILTERED",rgb_img_filtered); 
	cvNamedWindow("IR FILTERED",1); 
	cvShowImage("IR FILTERED",ir_img_filtered); 
	cvNamedWindow("DEPTH ORIGIN",1); 
	cvShowImage("DEPTH ORIGIN",depth_image); 
	imshow("disp8", disp8);
	imshow("rgb_reg", img2r);
 	imshow("ir_reg", img1r);
 	imshow("depth_reg", depth_image_dst);
	waitKey();*/

	//return disp8;
	return disp8_unrect;
}

//cv::Mat filter_op()
IplImage* filter_op()
{
	IplImage* rgb_img_r = cvCreateImage( cvSize(rgb_image->width, rgb_image->height), IPL_DEPTH_32F , 1 );
	IplImage* rgb_img_g = cvCreateImage( cvSize(rgb_image->width, rgb_image->height), IPL_DEPTH_32F , 1 );
	IplImage* rgb_img_b = cvCreateImage( cvSize(rgb_image->width, rgb_image->height), IPL_DEPTH_32F , 1 );
	rgb_img_filtered = cvCreateImage( cvSize(rgb_image->width, rgb_image->height), IPL_DEPTH_8U , 1 );
	for(int y = 0; y < rgb_image->height; y++){
		for(int x = 0; x < rgb_image->width; x++){
			CvScalar rgb_tmp = cvGet2D(rgb_image, y, x); 
			cvSet2D(rgb_img_r, y, x, cvScalar(rgb_tmp.val[2]) );
			cvSet2D(rgb_img_g, y, x, cvScalar(rgb_tmp.val[1]) );
			cvSet2D(rgb_img_b, y, x, cvScalar(rgb_tmp.val[0]) );
		}
	}
	CvMat* filter_r = cvCreateMat(3,3,CV_32FC1);
	cvSet2D(filter_r,0,0,cvRealScalar(0.1451)); cvSet2D(filter_r,0,1,cvRealScalar(0.1900));	cvSet2D(filter_r,0,2,cvRealScalar(0.1228));
	cvSet2D(filter_r,1,0,cvRealScalar(-0.0354));cvSet2D(filter_r,1,1,cvRealScalar(0.0089));	cvSet2D(filter_r,1,2,cvRealScalar(0.1244));
	cvSet2D(filter_r,2,0,cvRealScalar(0.1708)); cvSet2D(filter_r,2,1,cvRealScalar(0.0985));	cvSet2D(filter_r,2,2,cvRealScalar(0.1809));
	cvFilter2D(rgb_img_r,rgb_img_r,filter_r,cvPoint(-1,-1));
	CvMat* filter_g = cvCreateMat(3,3,CV_32FC1);
	cvSet2D(filter_g,0,0,cvRealScalar(0.1844));cvSet2D(filter_g,0,1,cvRealScalar(-0.0806));cvSet2D(filter_g,0,2,cvRealScalar(0.1249));
	cvSet2D(filter_g,1,0,cvRealScalar(0.1866));cvSet2D(filter_g,1,1,cvRealScalar(-0.1393));cvSet2D(filter_g,1,2,cvRealScalar(0.1129));
	cvSet2D(filter_g,2,0,cvRealScalar(0.1981));cvSet2D(filter_g,2,1,cvRealScalar(-0.0841));cvSet2D(filter_g,2,2,cvRealScalar(0.0841));
	cvFilter2D(rgb_img_g,rgb_img_g,filter_g,cvPoint(-1,-1));
	CvMat* filter_b = cvCreateMat(3,3,CV_32FC1);
	cvSet2D(filter_b,0,0,cvRealScalar(-0.0702));cvSet2D(filter_b,0,1,cvRealScalar(-0.0260));cvSet2D(filter_b,0,2,cvRealScalar(-0.0098));
	cvSet2D(filter_b,1,0,cvRealScalar(-0.1430));cvSet2D(filter_b,1,1,cvRealScalar(-0.0915));cvSet2D(filter_b,1,2,cvRealScalar(-0.0600));
	cvSet2D(filter_b,2,0,cvRealScalar(-0.0984));cvSet2D(filter_b,2,1,cvRealScalar(-0.0654));cvSet2D(filter_b,2,2,cvRealScalar(-0.0365));
	cvFilter2D(rgb_img_b,rgb_img_b,filter_b,cvPoint(-1,-1));

	float offset = -56.6978;
	//float offset = 0;
	for(int y = 0; y < rgb_image->height; y++){
		for(int x = 0; x < rgb_image->width; x++){
			CvScalar r_tmp = cvGet2D(rgb_img_r, y, x); CvScalar g_tmp = cvGet2D(rgb_img_g, y, x); CvScalar b_tmp = cvGet2D(rgb_img_b, y, x);
			cvSet2D(rgb_img_r, y, x, cvScalar(r_tmp.val[0]+g_tmp.val[0]+b_tmp.val[0]+offset) );
		}
	}
	cvConvertScale(rgb_img_r,rgb_img_filtered);

	CvMat* filter_ir = cvCreateMat(3,3,CV_32FC1);
	cvSet2D(filter_ir,0,0,cvRealScalar(-0.0049));cvSet2D(filter_ir,0,1,cvRealScalar(-0.0961));cvSet2D(filter_ir,0,2,cvRealScalar(0.0006));
	cvSet2D(filter_ir,1,0,cvRealScalar(-0.1532));cvSet2D(filter_ir,1,1,cvRealScalar(1.0000)); cvSet2D(filter_ir,1,2,cvRealScalar(-0.1084));
	cvSet2D(filter_ir,2,0,cvRealScalar(0.0018)); cvSet2D(filter_ir,2,1,cvRealScalar(-0.0741));cvSet2D(filter_ir,2,2,cvRealScalar(0.0062));
	
	IplImage* ir_img_tmp = cvCreateImage( cvSize(ir_image->width, ir_image->height), IPL_DEPTH_32F , 1 );	
	cvConvertScale(ir_image,ir_img_tmp);
	cvFilter2D(ir_img_tmp,ir_img_tmp,filter_ir,cvPoint(-1,-1));
	ir_img_filtered = cvCreateImage( cvSize(ir_image->width, ir_image->height), IPL_DEPTH_8U , 1 );
	cvConvertScale(ir_img_tmp,ir_img_filtered);

	//cvFilter2D(ir_image,ir_image,filter_ir,cvPoint(-1,-1));
	//ir_img_filtered = cvCreateImage( cvSize(ir_image->width, ir_image->height), IPL_DEPTH_8U , 1 );
	//cvConvertScale(ir_image,ir_img_filtered);
	cvReleaseImage(&rgb_img_r); cvReleaseImage(&rgb_img_g); cvReleaseImage(&rgb_img_b); 

	//cv::Mat disp8 = calculate_disparity(ir_img_filtered, rgb_img_filtered);
	//return disp8;
	IplImage* disp8_unrect = calculate_disparity(ir_img_filtered, rgb_img_filtered);
	return disp8_unrect;

}

int main(int argc, char** argv)
{
        ros::init(argc, argv, "cross_modal_kinect");
        ros::NodeHandle nh;
	calculate_transform();
        
        cross_modal_kinect cross_modal_kinect(nh);
        
	cross_modal_kinect.sub_ir.subscribe();		
	cross_modal_kinect.sub_ir_info.subscribe();
	cross_modal_kinect.sub_depth.subscribe();

        while(ros::ok())
        {       
		if(!cross_modal_kinect.cq_ir.empty()){
			cross_modal_kinect.cq_ir.callOne(ros::WallDuration());
        		cross_modal_kinect.cq_ir_info.callOne(ros::WallDuration());
			cross_modal_kinect.sub_ir.unsubscribe();
			cross_modal_kinect.sub_ir_info.unsubscribe();
			if(ir_image && rgb_image && depth_image){

				cv::Mat disp8 = filter_op();
				//pub_ir_, pub_rgb_, pub_ir_filtered_, pub_rgb_filtered_, pub_depth_, pub_depth_refined_, pub_disp_
				sensor_msgs::ImagePtr msg_ir = sensor_msgs::CvBridge::cvToImgMsg(ir_image, "mono16");
				cross_modal_kinect.pub_ir_.publish(msg_ir);
				sensor_msgs::ImagePtr msg_rgb = sensor_msgs::CvBridge::cvToImgMsg(rgb_image, "bgr8");
				cross_modal_kinect.pub_rgb_.publish(msg_rgb);
				sensor_msgs::ImagePtr msg_ir_filtered = sensor_msgs::CvBridge::cvToImgMsg(ir_img_filtered, "mono8");
				cross_modal_kinect.pub_ir_filtered_.publish(msg_ir_filtered);
				sensor_msgs::ImagePtr msg_rgb_filtered = sensor_msgs::CvBridge::cvToImgMsg(rgb_img_filtered, "mono8");
				cross_modal_kinect.pub_rgb_filtered_.publish(msg_rgb_filtered);
				sensor_msgs::ImagePtr msg_depth = sensor_msgs::CvBridge::cvToImgMsg(depth_image, "mono16");
				cross_modal_kinect.pub_depth_.publish(msg_depth);
				IplImage disp8_tmp = IplImage(disp8);
				sensor_msgs::ImagePtr msg_disp = sensor_msgs::CvBridge::cvToImgMsg(&disp8_tmp, "mono8");
				cross_modal_kinect.pub_disp_.publish(msg_disp);

				CvScalar qq;				
				qq = cvGet2D(Q, 2, 3); double f = qq.val[0];
				qq = cvGet2D(Q, 3, 2); double inv_Tx = -qq.val[0];
				double f_Tx = f/inv_Tx;
				double w[2] = {0.0228, 0.6033};
				for (int v = 0; v < depth_image->height; ++v){
					for (int u = 0; u < depth_image->width; ++u){
						CvScalar depth_data = cvGet2D(depth_image, v, u);
						unsigned short disp_point = disp8.at<uchar>(v, u);				
						if (depth_data.val[0] == 0 && disp_point != 0){
							cvSet2D(depth_image, v, u, cvRealScalar(((f_Tx/disp_point)*w[0]+w[1])*1000) );
							CvScalar tmp = cvGet2D(depth_image, v, u);
							//printf("@ %d, %d from 0 to --> %f (%f)\n", v, u, tmp.val[0], ((f_Tx/disp_point)*w[0]+w[1])*1000);
						}
					}
				}
				sensor_msgs::ImagePtr msg_depth_refined = sensor_msgs::CvBridge::cvToImgMsg(depth_image, "mono16");
				cross_modal_kinect.pub_depth_refined_.publish(msg_depth_refined);

				cvNamedWindow("REFINED DEPTH",1); 
				cvShowImage("REFINED DEPTH",depth_image); 
				char filename[255];
				sprintf(filename,"depth-%d-refined.raw",img_index-1);
				FILE *pFile=fopen(filename,"w");
        			fwrite (depth_image->imageData , sizeof(unsigned short) , 640*480 , pFile );
        			fclose (pFile);
				//waitKey();

				Eigen::Matrix4d depth_to_rgb;
				double d2r_tmp[3][4];
				for(int y = 0; y < 3; y++)
					for(int x = 0; x < 4; x++){
						CvScalar e = x<3 ? cvGet2D(R,y,x) : cvGet1D(T,y);
						d2r_tmp[y][x] = e.val[0];
					}
				depth_to_rgb << d2r_tmp[0][0], d2r_tmp[0][1], d2r_tmp[0][2], d2r_tmp[0][3], 
						d2r_tmp[1][0], d2r_tmp[1][1], d2r_tmp[1][2], d2r_tmp[1][3], 
						d2r_tmp[2][0], d2r_tmp[2][1], d2r_tmp[2][2], d2r_tmp[2][3], 
						0, 0, 0, 1;
				CvScalar tmp;				
				tmp = cvGet2D(M1, 0, 2); float depth_cx = tmp.val[0];
				tmp = cvGet2D(M1, 1, 2); float depth_cy = tmp.val[0];
				tmp = cvGet2D(M1, 0, 0); double inv_depth_fx = 1.0 / tmp.val[0];
				tmp = cvGet2D(M1, 1, 1); double inv_depth_fy = 1.0 / tmp.val[0];

				tmp = cvGet2D(M2, 0, 2); float rgb_cx = tmp.val[0];
				tmp = cvGet2D(M2, 1, 2); float rgb_cy = tmp.val[0];
				tmp = cvGet2D(M2, 0, 0); float rgb_fx = tmp.val[0];
				tmp = cvGet2D(M2, 1, 1); float rgb_fy = tmp.val[0];								
				float bad_point = std::numeric_limits<float>::quiet_NaN ();

				/*cv::Mat depth_image_src(depth_image,0);
				cv::Mat depth_image_dst;
				remap(depth_image_src, depth_image_dst, map11, map12, INTER_LINEAR);
				IplImage qImg = IplImage(depth_image_dst);
				char filename[255];
				sprintf(filename,"xxx_depth_reg_dst.raw");
				FILE *pFile=fopen(filename,"w");
        			fwrite (qImg.imageData , sizeof(unsigned short) , 640*480 , pFile );
        			fclose (pFile);*/

				reg_depth_image = cvCreateImage( cvSize(rgb_image->width, rgb_image->height), IPL_DEPTH_16U , 1 );
				cvSet(reg_depth_image, cvScalar(0));
				for (int v = 0; v < depth_image->height; ++v){
					for (int u = 0; u < depth_image->width; ++u){
						CvScalar depth_data = cvGet2D(depth_image,v,u);				
						double depth = double(depth_data.val[0]);
						Eigen::Vector4d xyz_depth;
						xyz_depth << ((u - depth_cx)*depth) * inv_depth_fx,
							((v - depth_cy)*depth) * inv_depth_fy, depth, 1;
						Eigen::Vector4d xyz_rgb = depth_to_rgb * xyz_depth;
						double inv_Z = 1.0 / xyz_rgb.z();
						int u_rgb = (rgb_fx*xyz_rgb.x())*inv_Z + rgb_cx + 0.5;
						int v_rgb = (rgb_fy*xyz_rgb.y())*inv_Z + rgb_cy + 0.5;
						if (u_rgb < 0 || u_rgb >= rgb_image->width || 
							v_rgb < 0 || v_rgb >= rgb_image->height)
							continue;
						CvScalar reg_depth = cvGet2D(reg_depth_image, v_rgb, u_rgb);
						uint16_t  new_depth = (xyz_rgb.z()) + 0.5;
						// Validity and Z-buffer checks
						if ( reg_depth.val[0] == 0 || reg_depth.val[0] > new_depth)
							cvSet2D(reg_depth_image, v_rgb, u_rgb, cvScalar(new_depth));
					}
				}

				tmp = cvGet2D(M2, 0, 2); float center_x = tmp.val[0];
				tmp = cvGet2D(M2, 1, 2); float center_y = tmp.val[0];
				tmp = cvGet2D(M2, 0, 0); float constant_x = 0.001 / tmp.val[0];
				tmp = cvGet2D(M2, 1, 1); float constant_y = 0.001 / tmp.val[0];
				PointCloud::Ptr cloud_msg_refined (new PointCloud);
				cloud_msg_refined->height = ir_image->height;
				cloud_msg_refined->width  = ir_image->width;
				cloud_msg_refined->is_dense = false;
				cloud_msg_refined->points.resize (cloud_msg_refined->height * cloud_msg_refined->width);	
				for (int v = 0; v < (int)cloud_msg_refined->height; v++){
					for (int u = 0; u < (int)cloud_msg_refined->width; u++){

						// Fill in color
						CvScalar rgb_buffer = cvGet2D(rgb_image, v, u);
						RGBValue color;
						color.Red   = rgb_buffer.val[2];
						color.Green = rgb_buffer.val[1];
						color.Blue  = rgb_buffer.val[0];
						cloud_msg_refined->points[v*(int)cloud_msg_refined->height+u].rgb = color.float_value;

						CvScalar depth_data = cvGet2D(reg_depth_image, v, u);
				
						if (depth_data.val[0] == 0){
							cloud_msg_refined->points[v*(int)cloud_msg_refined->height+u].x = bad_point;
							cloud_msg_refined->points[v*(int)cloud_msg_refined->height+u].y = bad_point;
							cloud_msg_refined->points[v*(int)cloud_msg_refined->height+u].z = bad_point;
						}
						else{
							cloud_msg_refined->points[v*(int)cloud_msg_refined->height+u].x =(u-center_x)*depth_data.val[0] * constant_x;
							cloud_msg_refined->points[v*(int)cloud_msg_refined->height+u].y =(v-center_y)*depth_data.val[0] * constant_y;
							cloud_msg_refined->points[v*(int)cloud_msg_refined->height+u].z = depth_data.val[0] * 0.001;
						}
      					}
    				}
				cloud_msg_refined->header.frame_id = "/cross_modal_kinect";
				cross_modal_kinect.pub_point_cloud_refined_.publish(cloud_msg_refined);

				ir_image = NULL;
				rgb_image = NULL;
				depth_image = NULL;		
			}
		}
		if(!cross_modal_kinect.cq_rgb.empty()){			
			cross_modal_kinect.cq_rgb.callOne(ros::WallDuration());
			cross_modal_kinect.cq_rgb_info.callOne(ros::WallDuration());
			cross_modal_kinect.cq_depth.callOne(ros::WallDuration());	
			cross_modal_kinect.cq_depth_info.callOne(ros::WallDuration());
			img_index = img_index + 1;
			cross_modal_kinect.sub_rgb.unsubscribe();
			cross_modal_kinect.sub_rgb_info.unsubscribe();
		}
		//ros::getGlobalCallbackQueue()->callAvailable(ros::WallDuration(0)); 								
	}
        return 0;
}
